#include "worker.h"
#include "TSystem.h"
#include <QFile>
#include <QString>
#include <QTextStream>
#include <QDebug>
#include <QDir>

bool Worker::processCoincidence(const std::vector<MyPulse>& pulses, Cone& newCone){
    if (pulses.size() != 2)
    {
        return false;
    }
    // energy cut
    double Etot = pulses[1].height + pulses[0].height ;
    if (config->energyCut)
    {
        if (Etot < config->energyLow || Etot > config->energyUp)
        {
            return false;
        }
    }
    if (pulses[0].height < 0.01 || pulses[1].height < 0.01)
        return false;
    double cosHalfangle = 1 - pulses[0].height * 0.511 / (pulses[1].height * Etot);
    if (std::abs(cosHalfangle)>=1)
    {
        return false;
    }
    Vector3D axis = pulses[1].pos - pulses[0].pos;
    if (axis * axis == 0){
        return false;
    }
    newCone = Cone(pulses[0].pos, axis, cosHalfangle,
                   Etot, pulses[0].height);
    return true;
}

Worker::Worker(QObject *parent, const Setup* config_, PlotData* plotdata_) :
    QThread(parent),
    config(config_),
    plotdata(plotdata_)
{
//    stopped=true;
    stopped = false;
    exitted=false;

    for (int i=0;i<config->channelSettings.size();i++)
    {
        if(config->channelSettings[i].enabled)
            enabledChs.append(config->channelSettings[i].chNum);
    }

    infile.setFileName(config->filePath);
    QString outfp = config->outputDir.absolutePath() + "/cones.txt";
//    qDebug()<< outfp << '\n';
    outfile.open(outfp.toStdString(), std::ios::out);
    if (!infile.open(QIODevice::ReadOnly | QIODevice::Text))
    {
        throw std::invalid_argument("Cannot open file: " + config->filePath.toStdString());
    }
    if (!outfile.good())
    {
        throw std::invalid_argument("Cannot open file: " + outfp.toStdString());
    }
}

void Worker::run()
{
    QString line;
    QTextStream in(&infile);
    // skip header (first line)
    in.readLineInto(&line);

    while(!exitted)
    {
        if (!stopped)
        {
            // read some cones
            std::vector<Cone> cones;
            if(config->inputFormat=="CSV"){
                readCSV(cones, in);
            }
            else if (config->inputFormat=="BIN"){
//                readBIN(cones, in);
            }

            if (exitted)
                break;
            // update image
            plotdata->updateImage(cones.cbegin(), cones.cend(), true);

            if (exitted || plotdata->counts >= config->maxN)
                break;
        }
    }
    // save image & spectrum before quitting
    qDebug() << "Save results before exiting worker thread";
    plotdata->saveOnExit();
    qDebug() << "Worker thread exited.";
}

void Worker::readCSV(std::vector<Cone> &cones, QTextStream& in)
{
    QString line;
    std::vector<ulong> histNums;
    static std::vector<MyPulse> pulses;
    static double prevTT=0;
    int i=0;
    while (!exitted && i < config->chuckSize && in.readLineInto(&line))
    {
        MyPulse newPulse = MyPulse(line);
        auto it = enabledChs.indexOf(newPulse.cellNo);
        if (it < 0){
            // cell not found
            // go to next line
            continue;
        }
        else {
            newPulse.pos.X = config->channelSettings[it].x;
            newPulse.pos.Y = config->channelSettings[it].y;
            newPulse.height *= config->channelSettings[it].caliCoef;
            // TO DO: calculate z
        }
        if (pulses.size()==0 || std::abs(newPulse.timeStamp - prevTT) < config->timeWindow){
            prevTT = newPulse.timeStamp;
            pulses.push_back(std::move(newPulse));
        }
        else {
            // process coincident events
            Cone newCone;
            if (processCoincidence(pulses, newCone)){
                histNums.push_back(pulses[0].histNo);
                cones.push_back(std::move(newCone));
                i++;
            }
            pulses.clear();
            prevTT = newPulse.timeStamp;
            pulses.push_back(std::move(newPulse));
        }
    }

    // save cones to file
    for (int i=0;i<cones.size();i++)
    {
        outfile << std::fixed << std::setprecision(2)
                << std::setw(8) << cones[i].apex.X // << "    "
                << std::setw(8) << cones[i].apex.Y // << "    "
                << std::setw(8) << cones[i].apex.Z // << "    "
                << std::setw(8) << cones[i].axis.X // << "    "
                << std::setw(8) << cones[i].axis.Y // << "    "
                << std::setw(8) << cones[i].axis.Z // << "    "
                << std::fixed << std::setprecision(8)
                << std::setw(13) << cones[i].cosHalfAngle
                << std::setw(13) << cones[i].E0
                << std::setw(13) << cones[i].Edpst
                << std::setw(13) << histNums[i] << '\n';
    }

}

void Worker::handleStart()
{
    stopped=false;
}

void Worker::handlePause()
{
    stopped=true;
}

void Worker::handleStop()
{
    exitted = true;
    qDebug() << "Exiting thread signal received.";
}
